/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <dlog.h>
#include <wayland-client.h>
#include <Ecore_Wayland.h>
#include <Elementary.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <aul_screen_connector.h>

#include "screen_connector_watcher.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER"

struct _screen_connector_watcher_h {
	char *appid;
	int pid;
	void (*update_cb)(struct tizen_remote_surface *trs, struct wl_buffer *buffer, uint32_t time, void *data);
	void (*missing_cb)(struct tizen_remote_surface *trs, void *data);
	void (*change_cb)(struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
			int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys, void *data);
	void *data;
	struct tizen_remote_surface *surface;
	unsigned int event_filter;
	int resource_id;
};

typedef struct _screen_connector_watcher_h *screen_connector_watcher_h;

static struct tizen_remote_surface_manager *__rsm;
static struct wayland_tbm_client *__tbm_client;
static struct wl_tbm *__tbm;
static bool __is_init;
static GList *__watcher_list;
static GList *__pending_watcher_list;
static aul_screen_viewer_handler __aul_cbs;

static void __buffer_updated(void *data, struct tizen_remote_surface *trs, struct wl_buffer *buffer, uint32_t time)
{
	screen_connector_watcher_h watcher_h = (screen_connector_watcher_h)data;

	if (watcher_h)
		watcher_h->update_cb(trs, buffer, time, watcher_h->data);
}

static void __buffer_missing(void *data, struct tizen_remote_surface *trs)
{
	screen_connector_watcher_h watcher_h = (screen_connector_watcher_h)data;

	if (watcher_h)
		watcher_h->missing_cb(trs, watcher_h->data);
}

static void __buffer_changed(void *data, struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
		int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys)
{
	screen_connector_watcher_h watcher_h = (screen_connector_watcher_h)data;

	if (watcher_h)
		watcher_h->change_cb(trs, type, tbm, img_file_fd, img_file_size, time, keys, watcher_h->data);
}

static void __input_event_filter(void *data, struct tizen_remote_surface *trs, uint32_t event_filter)
{
	screen_connector_watcher_h watcher_h = (screen_connector_watcher_h)data;

	if (watcher_h)
		watcher_h->event_filter = event_filter;

	LOGD("event filter updated:%d", event_filter);
}

static const struct tizen_remote_surface_listener __rs_listener = {
	__buffer_updated,
	__buffer_missing,
	__buffer_changed,
	__input_event_filter,
};

static void __app_removed(const char *appid, const int pid, void *data)
{
}

EXPORT_API void screen_connector_watcher_redirect_surface(screen_connector_watcher_h watcher_h)
{
	if (watcher_h == NULL)
		return;

	watcher_h->surface = tizen_remote_surface_manager_create_surface(__rsm, (uint32_t)watcher_h->resource_id, __tbm);
	if (watcher_h->surface == NULL)
		return;

	LOGD("resource_id : %d, (%p)", watcher_h->resource_id, watcher_h->surface);
	tizen_remote_surface_add_listener(watcher_h->surface, &__rs_listener, watcher_h);
	tizen_remote_surface_redirect(watcher_h->surface);
}

static screen_connector_watcher_h __create_watcher_handle_info(char *appid)
{
	screen_connector_watcher_h watcher_h;

	watcher_h = (screen_connector_watcher_h)calloc(1, sizeof(struct _screen_connector_watcher_h));
	if (watcher_h == NULL) {
		LOGE("watcher_h calloc fail");
		return NULL;
	}

	watcher_h->appid = strdup(appid);
	if (watcher_h->appid == NULL) {
		LOGE("out of memory");
		free(watcher_h);
		return NULL;
	}

	return watcher_h;
}

static void __app_added(const char *appid, const int pid, const unsigned int resource_id, void *data)
{
	screen_connector_watcher_h watcher_h = NULL;
	screen_connector_watcher_h cur_h;
	GList *watcher_list = __watcher_list;

	LOGD("__app_added pid: %d, resource_id: %d, appid: %s", pid, resource_id, appid);
	for (; watcher_list != NULL; watcher_list = watcher_list->next) {
		cur_h = (screen_connector_watcher_h)watcher_list->data;
		if (strcmp(cur_h->appid, appid) == 0) {
			watcher_h = cur_h;
			__watcher_list = g_list_remove(__watcher_list, cur_h);
			break;
		}
	}

	if (watcher_h == NULL) {
		watcher_h = __create_watcher_handle_info((char *)appid);
		watcher_h->pid = pid;
		watcher_h->resource_id = resource_id;
		__pending_watcher_list = g_list_append(__pending_watcher_list, watcher_h);

		LOGD("add pending list: %d, resource_id: %d, appid: %s", pid, resource_id, appid);
		return;
	}

	watcher_h->resource_id = resource_id;
	watcher_h->pid = pid;
	if (resource_id > 0)
		screen_connector_watcher_redirect_surface(watcher_h);
	else
		LOGE("bad tbm message received. missing arguments");
}

EXPORT_API int screen_connector_watcher_init(void)
{
	int ret = 0;
	struct wl_registry *registry;
	Ecore_Wl_Global *global;
	Eina_Inlist *globals;

	if (__is_init)
		return 0;

	ecore_wl_init(NULL);

	registry = ecore_wl_registry_get();
	globals = ecore_wl_globals_get();

	if (!registry || !globals) {
		LOGE("could not get registry(%p) or global list(%p)", registry,
				globals);
		return -1;
	}

	EINA_INLIST_FOREACH(globals, global) {
		if (!strcmp(global->interface, "tizen_remote_surface_manager")) {
			__rsm = wl_registry_bind(registry, global->id,
					&tizen_remote_surface_manager_interface,
					global->version < 4 ? global->version : 4);
		}
	}

	if (!__rsm) {
		LOGE("could not get remote surface manager");
		return -1;
	}

	__tbm_client = (struct wayland_tbm_client *)wayland_tbm_client_init(ecore_wl_display_get());
	if (!__tbm_client) {
		LOGE("could not get tbm client");
		return -1;
	}

	__tbm = (struct wl_tbm *)wayland_tbm_client_get_wl_tbm(__tbm_client);
	if (!__tbm) {
		LOGE("could not get tbm");
		return -1;
	}

	__aul_cbs.app_added = __app_added;
	__aul_cbs.app_removed = __app_removed;
	aul_screen_connector_add_screen_viewer(&__aul_cbs, NULL);

	__is_init = true;

	return ret;
}

EXPORT_API int screen_connector_watcher_fini(void)
{
	if (!__is_init)
		return 0;

	if (aul_screen_connector_remove_app_screen() != 0)
		return -1;

	if (__tbm_client)
		wayland_tbm_client_deinit(__tbm_client);

	if (__rsm)
		tizen_remote_surface_manager_destroy(__rsm);

	ecore_wl_shutdown();

	__is_init = false;
	return 0;
}

static void __destroy_watcher_h(screen_connector_watcher_h handle)
{
	if (!handle)
		return;

	if (handle->surface) {
		tizen_remote_surface_set_user_data(handle->surface, NULL);
		tizen_remote_surface_unredirect(handle->surface);
		tizen_remote_surface_destroy(handle->surface);
		handle->surface = NULL;
	}

	if (__pending_watcher_list)
		__pending_watcher_list = g_list_remove(__pending_watcher_list, handle);

	if (__watcher_list)
		__watcher_list = g_list_remove(__watcher_list, handle);

	if (handle->appid) {
		free(handle->appid);
		handle->appid = NULL;
	}

	free(handle);
}

EXPORT_API screen_connector_watcher_h screen_connector_watcher_create_handle(char *id, int pid, screen_connector_watcher_ops *ops, void *data)
{
	screen_connector_watcher_h watcher_h = NULL;

	watcher_h = __create_watcher_handle_info(id);
	if (watcher_h == NULL)
		return NULL;

	watcher_h->update_cb = ops->update_cb;
	watcher_h->change_cb = ops->change_cb;
	watcher_h->missing_cb = ops->missing_cb;
	watcher_h->pid = pid;
	watcher_h->data = data;

	return watcher_h;
}

EXPORT_API screen_connector_watcher_h screen_connector_watcher_add(screen_connector_watcher_ops *ops, char *id, void *data)
{
	screen_connector_watcher_h watcher_h = NULL;
	screen_connector_watcher_h cur_h;
	GList *watcher_list = __pending_watcher_list;

	for (; watcher_list != NULL; watcher_list = watcher_list->next) {
		cur_h = (screen_connector_watcher_h)watcher_list->data;
		if (strcmp(cur_h->appid, id) == 0) {
			watcher_h = cur_h;
			__pending_watcher_list = g_list_remove(__pending_watcher_list, cur_h);
			break;
		}
	}

	if (watcher_h != NULL) {
		watcher_h->update_cb = ops->update_cb;
		watcher_h->change_cb = ops->change_cb;
		watcher_h->missing_cb = ops->missing_cb;
		watcher_h->data = data;
		screen_connector_watcher_redirect_surface(watcher_h);
	} else {
		watcher_h = __create_watcher_handle_info(id);
		watcher_h->update_cb = ops->update_cb;
		watcher_h->change_cb = ops->change_cb;
		watcher_h->missing_cb = ops->missing_cb;
		watcher_h->data = data;
		__watcher_list = g_list_append(__watcher_list, watcher_h);
	}

	return watcher_h;
}

EXPORT_API int screen_connector_watcher_remove(screen_connector_watcher_h h)
{
	if (h == NULL) {
		LOGE("screen_connector_watcher_h is NULL");
		return -1;
	}
	__destroy_watcher_h(h);

	return 0;
}

EXPORT_API int screen_connector_watcher_update(const char *appid)
{
	return aul_screen_connector_update_app_screen(appid);
}

EXPORT_API char *screen_connector_watcher_get_appid(screen_connector_watcher_h h)
{
	if (h == NULL)
		return NULL;

	return h->appid;
}

EXPORT_API int screen_connector_watcher_get_pid(screen_connector_watcher_h h)
{
	if (h == NULL)
		return -1;

	return h->pid;
}

EXPORT_API int screen_connector_watcher_get_resource_id(screen_connector_watcher_h h)
{
	if (h == NULL)
		return -1;

	return h->resource_id;
}

EXPORT_API int screen_connector_watcher_set_resource_id(screen_connector_watcher_h h, int rid)
{
	if (h == NULL)
		return -1;

	h->resource_id = rid;
	return 0;
}

EXPORT_API struct tizen_remote_surface *screen_connector_watcher_get_trs(screen_connector_watcher_h h)
{
	if (h == NULL)
		return NULL;

	return h->surface;
}

EXPORT_API unsigned int screen_connector_watcher_get_event_filter(screen_connector_watcher_h h)
{
	if (h == NULL)
		return 0;

	return h->event_filter;
}

EXPORT_API int screen_connector_watcher_check_input_disabled(screen_connector_watcher_h h,
	screen_connector_input_type_e input)
{
	if (h == NULL) {
		LOGE("inavlid argument");
		return -1;
	}

	return ((h->event_filter & input) > 0);
}
